import logging
from pathlib import Path
import pickle
import signal
import threading
import time
from typing import Dict
from soin.job import JobStatus, Queue, SpiderJob
from soin.utils import while_true


class Spider:

    root: Path
    stop: bool
    failed_job_dir: Path
    dispatch_queue: Queue
    writeback_queue: Queue
    state: Dict

    def __init__(self):
        self.state_file = self.root / "spider_state.pkl"
        self.state_save_lock = threading.Lock()
        self.state_read_lock = threading.Lock()
        self.stop = False
        self.state = {}
        self.failed_job_dir = self.root / "failed_jobs"
        self.failed_job_dir.mkdir(parents=True, exist_ok=True)
        self.worker_execute_time = 0
        signal.signal(signal.SIGINT, self._handle_sig)
        signal.signal(signal.SIGTERM, self._handle_sig)
        self.setup()

    def main(self):
        try:
            self.load_state()
            self.spawn_cleanup_thread()
            self.spawn_worker_thread()
            self.run()
        finally:
            self.job_result_thread.join()
            self.job_cleanup_thread.join()
            self.cleanup()

    def run(self):
        raise NotImplementedError()

    def setup(self):
        pass

    def cleanup(self):
        self.save_state()

    def save_state(self):
        self.state_save_lock.acquire()
        self.state_file.write_bytes(pickle.dumps(self.state))
        self.state_save_lock.release()

    def load_state(self):
        self.state_read_lock.acquire()
        try:
            self.state = pickle.loads(self.state_file.read_bytes())
        except Exception as e:
            logging.warn(f"load state failed with {e}")
        self.state_read_lock.release()

    def dispatch(self, job: SpiderJob):
        self.dispatch_queue.push(job)

    def spawn_worker_thread(self):
        self.job_result_thread = threading.Thread(target=self._job_result_worker)
        self.job_result_thread.start()

    def spawn_cleanup_thread(self):
        self.job_cleanup_thread = threading.Thread(target=self._job_cleanup_worker)
        self.job_cleanup_thread.start()

    @while_true(10)
    def _job_cleanup_worker(self):
        logging.info("start to cleanup failed jobs")
        for failed_job_file in self.failed_job_dir.glob("*"):
            job = pickle.loads(failed_job_file.read_bytes())
            failed_job_file.unlink()
            logging.info(f"retry failed job {job}")
            self.dispatch(job)
            time.sleep(1)
        logging.info("cleanup done")

    @while_true(1)
    def _job_result_worker(self):
        """线程中运行"""
        job : SpiderJob = self.writeback_queue.pop()
        if job is None:
            return
        logging.info(f"get writeback job: {job}")
        if job.status == JobStatus.success:
            job.on_success()
        if job.status == JobStatus.failed:
            job.on_failed()
            logging.warn(f"job failed with message: {job.message}")
            self._save_failed_job(job)

    def _save_failed_job(self, job: SpiderJob):
        (self.failed_job_dir / job.id).write_bytes(pickle.dumps(job))

    def as_worker(self, max_step=None):
        execute_count = 0
        while True:
            time.sleep(1)
            execute_count += 1
            if self.stop:
                break
            if execute_count == max_step:
                self.stop = True
                logging.info(f"exceed max execution step {execute_count}, quit")
                break
            job : SpiderJob = self.dispatch_queue.pop()
            if job is None:
                continue
            logging.info(f"got job {job}")
            job.execute()
            self.writeback_queue.push(job)
            logging.info(f"{job} has been processed")

    def _handle_sig(self, *args):
        logging.info("about to quit")
        self.stop = True



from datetime import datetime, timedelta
import enum
import logging
import os
import pickle
import typing
from uuid import uuid4
import soin

import redis


class JobStatus(enum.Enum):
    success = 0
    failed = 1


class SpiderJob:

    # 爬虫
    if typing.TYPE_CHECKING:
        spider: soin.Spider

    # perform 完成后对该 Job 设置 result
    result: typing.Any

    # 任务状态
    status: JobStatus

    # 任务执行结果信息
    message: str

    # 任务运行时间
    execution_time: timedelta

    # 节点
    hostname: str

    def __init__(self):
        self.id = uuid4().hex
        self.result = None
        self.status = None
        self.message = None
        self.hostname = None
        self.execution_time = None

    def __repr__(self) -> str:
        return f"SpiderJob@{self.id} status:{self.status} message:{self.message} hostname:{self.hostname}"

    __str__ = __repr__

    def set_spider(self, spider: "soin.Spider"):
        self.spider = spider

    def execute(self):
        self.hostname = os.uname().nodename
        before_execute = datetime.now()
        self.status = JobStatus.success
        self.message = "ok"
        try:
            self.result = self.perform()
        except Exception as e:
            self.status = JobStatus.failed
            self.message = str(e)
        self.execution_time = datetime.now() - before_execute

    def perform(self):
        """运行在子节点 worker 中的逻辑
        """
        raise NotImplementedError()

    def on_failed(self):
        raise NotImplementedError()

    def on_success(self):
        """运行在主节点 worker 中的逻辑
        """
        raise NotImplementedError()


class Queue:

    def push(self):
        pass

    def pop(self):
        pass


class RedisQueue(Queue):

    def __init__(self, queue_name, redis: redis.Redis):
        self.queue_name = queue_name
        self.redis = redis

    def push(self, job: SpiderJob):
        job_bytes = pickle.dumps(job)
        self.redis.lpush(self.queue_name, job_bytes)

    def pop(self) -> SpiderJob:
        results = self.redis.brpop(self.queue_name, timeout=3)
        if not results:
            return None
        try:
            return pickle.loads(results[1])
        except Exception as e:
            logging.error(f"pickle job failed: {e}")
            return None



import functools
import logging
import os
import random
import shutil
import signal
import time
import requests
import tempfile
from pathlib import Path

from soin.job import SpiderJob


class TestJob(SpiderJob):

    def on_success(self):
        pass

    def on_failed(self):
        pass

    def perform(self):
        logging.info("test job is running")
        if random.random() < 0.3:
            logging.error("test job exception occued")
            raise Exception("test exception from test job")
        return 1


def while_true(interval=1):
    def decorator(func):
        @functools.wraps(func)
        def __inner(self, *args, **kwargs):
            while True:
                if self.stop:
                    break
                func(self, *args, **kwargs)
                time.sleep(interval)

        return __inner

    return decorator


def setup_root(root: str) -> Path:
    path = Path(root)
    path.mkdir(exist_ok=True, parents=True)
    return path


def setup_logger(log_file, level=logging.INFO):
    logging.basicConfig(
        filename=log_file,
        level=level,
        format="[%(asctime)s] %(levelname)s in %(module)s: %(message)s",
    )


def download_file(url: str, dest: str = None) -> bytes:
    fp = tempfile.NamedTemporaryFile(delete=False)
    for i in range(5):
        fp.seek(0)
        try:
            with requests.get(url, stream=True) as r:
                r.raise_for_status()
                for chunk in r.iter_content(chunk_size=1024**2):
                    fp.write(chunk)
            break
        except Exception as e:
            logging.warn(
                f"download file from {url} failed for {i+1} time, exception: {e}"
            )
    else:
        logging.error(
            f"download file from {url} failed after {i + 1}, quit"
        )
        fp.close()
        os.remove(fp.name)
        raise e
    if dest is None:
        fp.seek(0)
        bytes = fp.read()
        fp.close()
        os.remove(fp.name)
        return bytes
    else:
        fp.close()
        shutil.copy(fp.name, dest)
        os.remove(fp.name)
        return None


def retry(count: int):
    def decorator(func):
        @functools.wraps(func)
        def real_func(*args, **kwargs):
            for i in range(count):
                try:
                    res = func(*args, **kwargs)
                    break
                except Exception as e:
                    logging.exception(
                        f"call func {func.__name__} failed for {i} time, exception {e}"
                    )
                    time.sleep(1 + i)
            else:
                logging.error(
                    f"call func {func.__name__} failed after {count} time"
                )
                return None
            return res

        return real_func

    return decorator



from typing import Dict
import time
from uuid import uuid4
import requests
import logging


class Request:
    def __init__(
        self,
        url,
        method,
        headers,
        preset_body=None,
        preset_params=None,
        timeout=10,
        retry=5,
    ):
        self.id = uuid4().hex
        self.url = url
        self.method = method
        self.headers = headers
        self.timeout = timeout
        self.retry = retry
        self.preset_body = preset_body or {}
        self.preset_params = preset_params or {}

    def __str__(self):
        return f"url: {self.url} - method: {self.method}"

    __repr__ = __str__

    def make(self, data: Dict = None, params: Dict = None) -> requests.Response:
        logging.debug(f"make request {self} with data: {data}, params: {params}")
        response: requests.Response = None
        data = data or {}
        params = params or {}

        data = {**self.preset_body, **data}
        params = {**self.preset_params, **params}

        for i in range(self.retry):
            try:
                response = requests.request(
                    self.method,
                    self.url,
                    headers=self.headers,
                    params=params,
                    json=data,
                    timeout=self.timeout,
                )
            except Exception as e:
                logging.warn(
                    f"error while making request {self}, exception is {e}, retry {i}"
                )
                time.sleep(3)
                continue
            break
        else:
            logging.error(
                f"request failed after retrying"
            )
            return None

        return response



import importlib
import logging
import os
import sys
import click

import soin

quit = False

@click.group()
def cli():
    pass


@cli.command()
@click.argument("module")
@click.option("-w", "--worker", is_flag=True, default=False, help="spawn as worker")
@click.option("-s", "--steps", default=None, help="max step for a worker to run")
def start(module, worker, steps=None):
    sys.path.insert(0, os.getcwd())
    module = importlib.import_module(module)
    spider : soin.Spider = module.Main()
    if worker:
        spider.as_worker(steps)
    else:
        spider.main()
    logging.info("Gracefully existed!")


if __name__ == "__main__":
    cli()



from setuptools import setup, find_packages

setup(
    name='soin',
    version='1.0.4',
    url='https://github.com/louis-she/soin.git',
    author='Chenglu',
    author_email='chenglu.she@gmail.com',
    description='tool set for spiders',
    long_description="tool set for spiders",
    packages=find_packages(),
    entry_points={
        'console_scripts': [
            'soin = soin.cli:cli',
        ],
    }
)
